import random
import timeit

def merge(left, right):
    """assume left and right are already sorted"""
    left_finger = 0
    right_finger = 0
    merged = []

    while left_finger < len(left) and right_finger < len(right):
        if left[left_finger] <= right[right_finger]:
            merged.append(left[left_finger])
            left_finger += 1
        else:
            merged.append(right[right_finger])
            right_finger += 1

    # we've gone through all the items from one list
    # just append everything from the remaining list
    if left_finger == len(left):
        # append the right
        # for i in range(right_finger, len(right)):
        #     merged.append(right[right_finger])
        # merged.extend(right[right_finger:])
        merged += right[right_finger:]
    else:
        # append the left
        merged += left[left_finger:]

    return merged

def merge_sort(arr):
    # base case: nothing to sort for a list of 1 or 0 elements
    if len(arr) <= 1:
        return arr

    # find an index to split arr in half
    midpoint = int(len(arr) / 2)

    left = merge_sort(arr[:midpoint])
    right = merge_sort(arr[midpoint:])
    return merge(left, right)

def insertion_sort(arr):
    for i in range(1, len(arr)):
        tmp = arr[i]
        j = i
        while j > 0 and tmp < arr[j - 1]:
            arr[j] = arr[j - 1]
            j -= 1
        arr[j] = tmp

# check for correctness
arr = [random.random() for _ in range(1000)]
insertion_sort(arr)
assert all(arr[i] >= arr[i - 1] for i in range(1, len(arr)))
arr = [random.random() for _ in range(1000)]
arr = merge_sort(arr)
assert all(arr[i] >= arr[i - 1] for i in range(1, len(arr)))

# timing
for n in [100, 1000, 2000, 5000, 10000, 20000]:
    arr = [random.random() for _ in range(n)]
    print("n =", n)
    insertion_time = timeit.timeit("insertion_sort(arr[:])", number=5, globals=globals())
    merge_time = timeit.timeit("merge_sort(arr[:])", number=5, globals=globals())
    print("insertion sort: {:.6f}s".format(insertion_time))
    print("merge sort: {:.6f}s".format(merge_time))
    print("ratio of insertion sort to merge sort: {:.3f}".format(insertion_time / merge_time))
    print()
